from enum import Enum, EnumMeta

from flask_admin.form import Select2Field


class EnumField:
    @staticmethod
    def get_name(enum_instance):
        return enum_instance.name

    @staticmethod
    def get_value(enum_instance):
        return enum_instance.value

    default_label_function = get_value

    def apply_choices(
        self, enum_class: EnumMeta, label_function=default_label_function, kwargs=None
    ):
        if not kwargs:
            kwargs = dict()
        choices = [
            (member.value, label_function(member))
            for name, member in enum_class.__members__.items()
        ]
        kwargs["choices"] = choices
        kwargs["coerce"] = self.coerce
        return kwargs

    @staticmethod
    def coerce(data):
        if isinstance(data, Enum):
            return data.value
        return str(data)


class SelectEnumField(Select2Field, EnumField):
    def __init__(
        self, *args, enum_class: EnumMeta, label_function=EnumField.default_label_function, **kwargs
    ):
        kwargs = self.apply_choices(enum_class, label_function, kwargs)
        super().__init__(*args, **kwargs)



from pathlib import Path

from click.testing import CliRunner

from cg.apps.cgstats.crud import find
from cg.apps.cgstats.stats import StatsAPI
from cg.cli.demultiplex.add import select_project_cmd
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCell


def test_select_command(
    cli_runner: CliRunner,
    populated_stats_api: StatsAPI,
    demultiplexed_runs: Path,
    flow_cell: FlowCell,
    demultiplex_context: CGConfig,
):
    demultiplex_context.cg_stats_api_ = populated_stats_api
    # GIVEN a stats api with some information about a flowcell
    flowcell_id: str = flow_cell.id
    full_flow_cell_name: str = flow_cell.full_name
    assert populated_stats_api.find_handler.get_flowcell_id(flowcell_id)
    demux_results = DemuxResults(
        demux_dir=demultiplexed_runs / full_flow_cell_name,
        flow_cell=flow_cell,
        bcl_converter="bcl2fastq",
    )

    # GIVEN a project id
    project_id: str = ""
    for project in demux_results.projects:
        project_id = project.split("_")[-1]
    assert project_id

    # WHEN exporting the sample information
    result = cli_runner.invoke(
        select_project_cmd, [flowcell_id, "--project", project_id], obj=demultiplex_context
    )

    # THEN assert that the command exits with success
    assert result.exit_code == 0
    # THEN assert that the flowcell id if in the printing
    assert flowcell_id in result.output



import logging
from typing import List, Optional

import alchy
from sqlalchemy import or_

from cg.apps.cgstats.db.models import (
    Supportparams,
    Datasource,
    Demux,
    Flowcell,
    Project,
    Sample,
    Unaligned,
)
from cg.models.cgstats.stats_sample import StatsSample
from cg.models.demultiplex.demux_results import DemuxResults

LOG = logging.getLogger(__name__)
SAMPLE_PATTERN = "{}\_%"


class FindHandler:
    def get_support_parameters_id(self, demux_results: DemuxResults) -> Optional[int]:
        """Fetch the id of the support parameters if post exists"""
        LOG.debug("Search for support parameters with file %s", demux_results.results_dir)
        support_parameters_id: Optional[int] = Supportparams.exists(str(demux_results.results_dir))
        if support_parameters_id:
            LOG.debug("Found support parameters with id %s", support_parameters_id)
            return support_parameters_id
        LOG.debug("Could not find support parameters")
        return None

    def get_datasource_id(self, demux_results: DemuxResults) -> Optional[int]:
        """Fetch the datasource id for a certain run"""
        stats_path = {
            "bcl2fastq": demux_results.conversion_stats_path,
            "dragen": demux_results.demux_stats_path,
        }
        LOG.debug("Search for datasource with file %s", stats_path[demux_results.bcl_converter])
        datasource_id: Optional[int] = Datasource.exists(
            str(stats_path[demux_results.bcl_converter])
        )
        if datasource_id:
            LOG.debug("Found datasource with id %s", datasource_id)
            return datasource_id
        LOG.debug("Could not find datasource")
        return None

    def get_flowcell_id(self, flowcell_name: str) -> Optional[int]:
        LOG.debug("Search for flowcell %s", flowcell_name)
        flowcell_id: Optional[int] = Flowcell.exists(flowcell_name)
        if flowcell_id:
            LOG.debug("Found flowcell with id %s", flowcell_id)
            return flowcell_id
        LOG.debug("Could not find flowcell")
        return None

    def get_demux_id(self, flowcell_object_id: int, base_mask: str = "") -> Optional[int]:
        """Flowcell object id refers to a database object"""
        demux_id: Optional[int] = Demux.exists(flowcell_id=flowcell_object_id, basemask=base_mask)
        if demux_id:
            return demux_id
        return None

    def get_project_id(self, project_name: str) -> Optional[int]:
        project_id: Optional[int] = Project.exists(project_name=project_name)
        if project_id:
            return project_id
        return None

    def get_sample(self, sample_id: str):
        """Get a unique demux sample."""
        pattern = SAMPLE_PATTERN.format(sample_id)
        return Sample.query.filter(
            or_(Sample.samplename.like(pattern), Sample.samplename == sample_id)
        ).first()

    def get_sample_id(self, sample_id: str, barcode: str) -> Optional[int]:
        sample_id: Optional[int] = Sample.exists(sample_name=sample_id, barcode=barcode)
        if sample_id:
            return sample_id
        return None

    def get_unaligned_id(self, sample_id: int, demux_id: int, lane: int) -> Optional[int]:
        unaligned_id: Optional[int] = Unaligned.exists(
            sample_id=sample_id, demux_id=demux_id, lane=lane
        )
        if unaligned_id:
            return unaligned_id
        return None

    def get_samples(self, flowcell: str, project_name: Optional[str] = None) -> alchy.Query:
        query: alchy.Query = Sample.query.join(Sample.unaligned, Unaligned.demux, Demux.flowcell)
        if project_name:
            query = query.join(Sample.project).filter(Project.projectname == project_name)

        query = query.filter(Flowcell.flowcellname == flowcell)

        return query

    def project_sample_stats(
        self, flowcell: str, project_name: Optional[str] = None
    ) -> List[StatsSample]:
        samples_query: alchy.Query = self.get_samples(flowcell=flowcell, project_name=project_name)
        db_sample: Sample
        return [StatsSample.from_orm(db_sample) for db_sample in samples_query]



"""This api should handle everything around demultiplexing."""
import logging
from pathlib import Path
from typing import Dict, List, Optional

from typing_extensions import Literal

from cgmodels.trailblazer.constants import AnalysisTypes
from cg.apps.demultiplex.sbatch import DEMULTIPLEX_COMMAND, DEMULTIPLEX_ERROR
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles, BclConverter
from cg.constants.priority import SlurmQos
from cg.io.controller import WriteFile
from cg.models.demultiplex.flow_cell import FlowCell
from cg.models.demultiplex.sbatch import SbatchCommand, SbatchError
from cg.models.slurm.sbatch import Sbatch, SbatchDragen
from cgmodels.cg.constants import Pipeline

LOG = logging.getLogger(__name__)


class DemultiplexingAPI:
    """Demultiplexing API should deal with anything related to demultiplexing.

    This includes starting demultiplexing, creating sample sheets, creating base masks,
    """

    def __init__(self, config: dict, out_dir: Optional[Path] = None):
        self.slurm_api = SlurmAPI()
        self.slurm_account: str = config["demultiplex"]["slurm"]["account"]
        self.mail: str = config["demultiplex"]["slurm"]["mail_user"]
        self.run_dir: Path = Path(config["demultiplex"]["run_dir"])
        self.out_dir: Path = out_dir or Path(config["demultiplex"]["out_dir"])
        self.environment: str = config.get("environment", "stage")
        LOG.info("Set environment to %s", self.environment)
        self.dry_run: bool = False

    @property
    def slurm_quality_of_service(self) -> Literal[SlurmQos.HIGH, SlurmQos.LOW]:
        """Return SLURM quality of service."""
        return SlurmQos.LOW if self.environment == "stage" else SlurmQos.HIGH

    def set_dry_run(self, dry_run: bool) -> None:
        """Set dry run."""
        LOG.debug(f"Set dry run to {dry_run}")
        self.dry_run = dry_run
        self.slurm_api.set_dry_run(dry_run=dry_run)

    @staticmethod
    def get_sbatch_error(
        flow_cell: FlowCell,
        email: str,
        demux_dir: Path,
    ) -> str:
        """Create the sbatch error string."""
        LOG.info("Creating the sbatch error string")
        error_parameters: SbatchError = SbatchError(
            flow_cell_id=flow_cell.id,
            email=email,
            logfile=DemultiplexingAPI.get_stderr_logfile(flow_cell=flow_cell).as_posix(),
            demux_dir=demux_dir.as_posix(),
            demux_started=flow_cell.demultiplexing_started_path.as_posix(),
        )
        return DEMULTIPLEX_ERROR.format(**error_parameters.dict())

    @staticmethod
    def get_sbatch_command(
        run_dir: Path,
        unaligned_dir: Path,
        sample_sheet: Path,
        demux_completed: Path,
        flow_cell: FlowCell,
        environment: Literal["production", "stage"] = "stage",
    ) -> str:
        """Return sbatch command."""
        LOG.info("Creating the sbatch command string")
        command_parameters: SbatchCommand = SbatchCommand(
            run_dir=run_dir.as_posix(),
            demux_dir=unaligned_dir.parent.as_posix(),
            unaligned_dir=unaligned_dir.as_posix(),
            sample_sheet=sample_sheet.as_posix(),
            demux_completed_file=demux_completed.as_posix(),
            environment=environment,
        )
        return DEMULTIPLEX_COMMAND[flow_cell.bcl_converter].format(**command_parameters.dict())

    @staticmethod
    def demultiplex_sbatch_path(flow_cell: FlowCell) -> Path:
        """Get the path to where sbatch script file should be kept."""
        return Path(flow_cell.path, "demux-novaseq.sh")

    @staticmethod
    def get_run_name(flow_cell: FlowCell) -> str:
        """Create the run name for the sbatch job."""
        return f"{flow_cell.id}_demultiplex"

    @staticmethod
    def get_stderr_logfile(flow_cell: FlowCell) -> Path:
        """Create the path to the stderr logfile."""
        return Path(flow_cell.path, f"{DemultiplexingAPI.get_run_name(flow_cell)}.stderr")

    @staticmethod
    def get_stdout_logfile(flow_cell: FlowCell) -> Path:
        """Create the path to the stdout logfile."""
        return Path(flow_cell.path, f"{DemultiplexingAPI.get_run_name(flow_cell)}.stdout")

    def get_all_demultiplexed_flow_cell_dirs(self) -> List[Path]:
        """Return all demultiplex flow cell out directories."""
        demultiplex_flow_cells: List[Path] = []
        for flow_cell_dir in self.out_dir.iterdir():
            if flow_cell_dir.is_dir():
                LOG.debug(f"Found directory {flow_cell_dir}")
                demultiplex_flow_cells.append(flow_cell_dir)
        return demultiplex_flow_cells

    def flow_cell_out_dir_path(self, flow_cell: FlowCell) -> Path:
        """Create the path to where the demuliplexed result should be produced."""
        return Path(self.out_dir, flow_cell.path.name)

    def unaligned_dir_path(self, flow_cell: FlowCell) -> Path:
        """Create the path to where the demuliplexed result should be produced."""
        return Path(
            self.flow_cell_out_dir_path(flow_cell), DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME
        )

    def demultiplexing_completed_path(self, flow_cell: FlowCell) -> Path:
        """Return the path to demultiplexing complete file."""
        LOG.info(
            Path(self.flow_cell_out_dir_path(flow_cell), DemultiplexingDirsAndFiles.DEMUX_COMPLETE)
        )
        return Path(
            self.flow_cell_out_dir_path(flow_cell), DemultiplexingDirsAndFiles.DEMUX_COMPLETE
        )

    def is_demultiplexing_completed(self, flow_cell: FlowCell) -> bool:
        """Check the path to where the demuliplexed result should be produced."""
        LOG.info(f"Check if demultiplexing is ready for {flow_cell.path}")
        logfile: Path = self.get_stderr_logfile(flow_cell)
        if not logfile.exists():
            LOG.warning("Could not find logfile!")
            return False
        return self.demultiplexing_completed_path(flow_cell).exists()

    def is_demultiplexing_possible(self, flow_cell: FlowCell) -> bool:
        """Check if it is possible to start demultiplexing.

        This means that
            - flow cell should be ready for demultiplexing (all files in place)
            - sample sheet needs to exist
            - demultiplexing should not be running
        """
        LOG.info(f"Check if demultiplexing is possible for {flow_cell.id}")
        demultiplexing_possible = True
        if not flow_cell.is_flow_cell_ready():
            demultiplexing_possible = False

        if not flow_cell.sample_sheet_exists():
            LOG.warning(f"Could not find sample sheet for {flow_cell.id}")
            demultiplexing_possible = False

        if flow_cell.is_demultiplexing_started():
            LOG.warning("Demultiplexing has already been started")
            demultiplexing_possible = False

        if self.flow_cell_out_dir_path(flow_cell=flow_cell).exists():
            LOG.warning("Flow cell out dir exists")
            demultiplexing_possible = False

        if self.is_demultiplexing_completed(flow_cell):
            LOG.warning(f"Demultiplexing is already completed for flow cell {flow_cell.id}")
            demultiplexing_possible = False
        return demultiplexing_possible

    def create_demultiplexing_started_file(self, demultiplexing_started_path: Path) -> None:
        LOG.info("Creating demultiplexing started file")
        if self.dry_run:
            return
        demultiplexing_started_path.touch(exist_ok=False)

    @staticmethod
    def get_trailblazer_config(slurm_job_id: int) -> Dict[str, List[str]]:
        return {"jobs": [str(slurm_job_id)]}

    @staticmethod
    def write_trailblazer_config(content: dict, file_path: Path) -> None:
        """Write the content to a yaml file"""
        LOG.info("Writing yaml content %s to %s", content, file_path)
        WriteFile.write_file_from_content(
            content=content, file_format=FileFormat.YAML, file_path=file_path
        )

    def add_to_trailblazer(self, tb_api: TrailblazerAPI, slurm_job_id: int, flow_cell: FlowCell):
        """Add demultiplexing entry to trailblazer."""
        if self.dry_run:
            return
        self.write_trailblazer_config(
            content=self.get_trailblazer_config(slurm_job_id=slurm_job_id),
            file_path=flow_cell.trailblazer_config_path,
        )
        tb_api.add_pending_analysis(
            case_id=flow_cell.id,
            analysis_type=AnalysisTypes.OTHER,
            config_path=flow_cell.trailblazer_config_path.as_posix(),
            out_dir=flow_cell.trailblazer_config_path.parent.as_posix(),
            slurm_quality_of_service=self.slurm_quality_of_service,
            email=self.mail,
            data_analysis=str(Pipeline.DEMULTIPLEX),
        )

    def start_demultiplexing(self, flow_cell: FlowCell):
        """Start demultiplexing for a flow cell."""
        self.create_demultiplexing_started_file(flow_cell.demultiplexing_started_path)
        demux_dir: Path = self.flow_cell_out_dir_path(flow_cell=flow_cell)
        unaligned_dir: Path = self.unaligned_dir_path(flow_cell=flow_cell)
        LOG.info(f"Demultiplexing to {unaligned_dir}")
        if not self.dry_run:
            LOG.info(f"Creating demux dir {unaligned_dir}")
            unaligned_dir.mkdir(exist_ok=False, parents=True)

        log_path: Path = self.get_stderr_logfile(flow_cell=flow_cell)
        error_function: str = self.get_sbatch_error(
            flow_cell=flow_cell, email=self.mail, demux_dir=demux_dir
        )
        commands: str = self.get_sbatch_command(
            run_dir=flow_cell.path,
            unaligned_dir=unaligned_dir,
            sample_sheet=flow_cell.sample_sheet_path,
            demux_completed=self.demultiplexing_completed_path(flow_cell=flow_cell),
            flow_cell=flow_cell,
            environment=self.environment,
        )

        if flow_cell.bcl_converter == BclConverter.BCL2FASTQ:
            sbatch_parameters: Sbatch = Sbatch(
                account=self.slurm_account,
                commands=commands,
                email=self.mail,
                error=error_function,
                hours=36,
                job_name=self.get_run_name(flow_cell),
                log_dir=log_path.parent.as_posix(),
                memory=125,
                number_tasks=18,
                quality_of_service=self.slurm_quality_of_service,
            )
        if flow_cell.bcl_converter == BclConverter.DRAGEN:
            sbatch_parameters: SbatchDragen = SbatchDragen(
                account=self.slurm_account,
                commands=commands,
                email=self.mail,
                error=error_function,
                hours=36,
                job_name=self.get_run_name(flow_cell),
                log_dir=log_path.parent.as_posix(),
                quality_of_service=self.slurm_quality_of_service,
            )

        sbatch_content: str = self.slurm_api.generate_sbatch_content(
            sbatch_parameters=sbatch_parameters
        )
        sbatch_path: Path = self.demultiplex_sbatch_path(flow_cell=flow_cell)
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info(f"Demultiplexing running as job {sbatch_number}")
        return sbatch_number



"""Enum classes has string or list values that behaves like defined """
from enum import Enum


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return str.__str__(self)


class ListEnum(list, Enum):
    pass


class IntEnum(int, Enum):
    def __int__(self) -> int:
        return int.__int__(self)



from typing import List
from cg.store import Store
from cg.store.models import Pool, Customer


def test_get_pools_to_invoice_query(
    store_with_pools_for_multiple_customers: Store,
    three_pool_names: List[str],
):
    """Test return pools to invoice."""
    # GIVEN a store with multiple pools for multiple customers
    assert len(store_with_pools_for_multiple_customers.get_pools()) > 1

    # WHEN finding samples to invoice
    pools: List[Pool] = store_with_pools_for_multiple_customers.get_pools_to_invoice_query().all()

    # THEN it should return all pools that are not invoiced
    for pool in pools:
        assert pool.name in three_pool_names
        assert isinstance(pool, Pool)


def test_get_pools_to_invoice_for_customer(
    store_with_pools_for_multiple_customers: Store,
    three_customer_ids: List[str],
):
    """Test return pools to invoice for a customer."""
    # GIVEN a store with multiple pools for multiple customers
    assert len(store_with_pools_for_multiple_customers.get_pools()) > 1

    # THEN the one customer can be retrieved
    customer: Customer = store_with_pools_for_multiple_customers.get_customer_by_internal_id(
        customer_internal_id=three_customer_ids[1]
    )
    assert customer

    # WHEN finding samples to invoice for a customer
    pools: List[Pool] = store_with_pools_for_multiple_customers.get_pools_to_invoice_for_customer(
        customer=customer
    )

    # THEN it should return all pools that are not invoiced for the customer
    assert pools[0].customer == customer



from typing import Optional, Any, List

from cg.constants import DataDelivery, Pipeline
from cg.models.orders.sample_base import OrderSample
from pydantic import constr, validator


class JsonSample(OrderSample):
    cohorts: Optional[List[str]]
    concentration: Optional[str]
    concentration_sample: Optional[str]
    control: Optional[str]
    data_analysis: Pipeline = Pipeline.MIP_DNA
    data_delivery: DataDelivery = DataDelivery.SCOUT
    index: Optional[str]
    panels: Optional[List[str]]
    quantity: Optional[str]
    synopsis: Optional[str]
    well_position: Optional[constr(regex=r"[A-H]:[0-9]+")]

    @validator("synopsis", pre=True)
    def join_list(cls, value: Any):
        if isinstance(value, list):
            return "".join(value)
        return value

    @validator("well_position", pre=True)
    def convert_well(cls, value: str):
        if not value:
            return None
        if ":" in value:
            return value
        return ":".join([value[0], value[1:]])



from typing import List, Union

from sqlalchemy.orm import Query
from cgmodels.cg.constants import Pipeline
from datetime import datetime

from cg.constants.constants import CaseActions, DataDelivery
from cg.constants.sequencing import SequencingMethod
from cg.constants.subject import PhenotypeStatus
from cg.store import Store
from cg.store.models import Analysis, Family, Sample
from cg.store.filters.status_case_filters import (
    filter_cases_by_case_search,
    filter_cases_by_customer_entry_ids,
    filter_cases_by_entry_id,
    filter_case_by_internal_id,
    filter_cases_by_name,
    filter_cases_by_pipeline_search,
    filter_cases_by_priority,
    filter_cases_not_analysed,
    get_newer_cases_by_creation_date,
    get_newer_cases_by_order_date,
    get_running_cases,
    filter_cases_by_ticket_id,
    get_cases_with_pipeline,
    get_cases_has_sequence,
    get_cases_for_analysis,
    get_cases_with_scout_data_delivery,
    get_report_supported_data_delivery_cases,
    get_cases_with_loqusdb_supported_pipeline,
    get_cases_with_loqusdb_supported_sequencing_method,
    get_inactive_analysis_cases,
    get_older_cases_by_creation_date,
)
from tests.store_helpers import StoreHelpers


def test_get_cases_has_sequence(base_store: Store, helpers: StoreHelpers, timestamp_now: datetime):
    """Test that a case is returned when there is a cases with a sequenced sample."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_now)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_get_cases_has_sequence_when_external(base_store: Store, helpers: StoreHelpers):
    """Test that a case is returned when there is a case with an externally sequenced sample."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=None, is_external=True)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_get_cases_has_sequence_when_not_sequenced(base_store: Store, helpers: StoreHelpers):
    """Test that no case is returned when there is a cases with sample that has not been sequenced."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=None)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_get_cases_has_sequence_when_not_external_nor_sequenced(
    base_store: Store, helpers: StoreHelpers
):
    """Test that no case is returned when there is a cases with sample that has not been sequenced nor is external."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=None, is_external=False)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_get_cases_with_pipeline_when_correct_pipline(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that no case is returned when there are no cases with the  specified pipeline."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_now)

    # GIVEN a cancer case
    test_case = helpers.add_case(base_store, data_analysis=Pipeline.BALSAMIC)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse for another pipeline
    cases: List[Query] = list(get_cases_with_pipeline(cases=cases, pipeline=Pipeline.BALSAMIC))

    # THEN cases should contain the test case
    assert cases


def test_get_cases_with_pipeline_when_incorrect_pipline(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that no case is returned when there are no cases with the  specified pipeline."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_now)

    # GIVEN a cancer case
    test_case: Family = helpers.add_case(base_store, data_analysis=Pipeline.BALSAMIC)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse for another pipeline
    cases: List[Query] = list(get_cases_with_pipeline(cases=cases, pipeline=Pipeline.MIP_DNA))

    # THEN cases should not contain the test case
    assert not cases


def test_get_cases_with_loqusdb_supported_pipeline(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test retrieval of cases that support Loqusdb upload."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_now)

    # GIVEN a MIP-DNA and a FLUFFY case
    test_mip_case: Family = helpers.add_case(base_store, data_analysis=Pipeline.MIP_DNA)
    test_mip_case.customer.loqus_upload = True
    test_fluffy_case: Family = helpers.add_case(
        base_store, name="test", data_analysis=Pipeline.FLUFFY
    )
    test_fluffy_case.customer.loqus_upload = True

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_mip_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.relate_sample(test_fluffy_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases with pipeline
    cases: List[Query] = list(get_cases_with_loqusdb_supported_pipeline(cases=cases))

    # THEN only the Loqusdb supported case should be extracted
    assert test_mip_case in cases
    assert test_fluffy_case not in cases


def test_get_cases_with_loqusdb_supported_sequencing_method(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test retrieval of cases with a valid Loqusdb sequencing method."""

    # GIVEN a sample with a valid Loqusdb sequencing method
    test_sample_wes: Sample = helpers.add_sample(
        base_store, sequenced_at=timestamp_now, application_type=SequencingMethod.WES
    )

    # GIVEN a MIP-DNA associated test case
    test_case_wes: Family = helpers.add_case(base_store, data_analysis=Pipeline.MIP_DNA)
    base_store.relate_sample(test_case_wes, test_sample_wes, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN retrieving the available cases
    cases: Query = get_cases_with_loqusdb_supported_sequencing_method(
        cases=cases, pipeline=Pipeline.MIP_DNA
    )

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN the expected case should be retrieved
    assert test_case_wes in cases


def test_get_cases_with_loqusdb_supported_sequencing_method_empty(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test retrieval of cases with a valid Loqusdb sequencing method."""

    # GIVEN a not supported loqusdb sample
    test_sample_wts: Sample = helpers.add_sample(
        base_store, name="sample_wts", sequenced_at=timestamp_now, is_rna=True
    )

    # GIVEN a MIP-DNA associated test case
    test_case_wts: Family = helpers.add_case(base_store, data_analysis=Pipeline.MIP_DNA)
    base_store.relate_sample(test_case_wts, test_sample_wts, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN retrieving the valid cases
    cases: Query = get_cases_with_loqusdb_supported_sequencing_method(
        cases=cases, pipeline=Pipeline.MIP_DNA
    )

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN no cases should be returned
    assert not cases.all()


def test_get_cases_for_analysis(base_store: Store, helpers: StoreHelpers, timestamp_now: datetime):
    """Test that a case is returned when there is a cases with an action set to analyse."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_now)

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(
        base_store, completed_at=timestamp_now, pipeline=Pipeline.MIP_DNA
    )

    # Given an action set to analyze
    test_analysis.family.action: str = CaseActions.ANALYZE

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_analysis.family, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_get_cases_for_analysis_when_sequenced_sample_and_no_analysis(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned when there are internally created cases with no action set and no prior analysis."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(
        base_store, sequenced_at=timestamp_now, is_external=False
    )

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_get_cases_for_analysis_when_cases_with_no_action_and_new_sequence_data(
    base_store: Store,
    helpers: StoreHelpers,
    timestamp_yesterday: datetime,
    timestamp_now: datetime,
):
    """Test that a case is returned when cases with no action, but new sequence data."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(
        base_store, sequenced_at=timestamp_now, is_external=False
    )

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(base_store, pipeline=Pipeline.MIP_DNA)

    # Given an action set to None
    test_analysis.family.action: Union[None, str] = None

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_analysis.family, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN an old analysis
    test_analysis.created_at = timestamp_yesterday

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_get_cases_for_analysis_when_cases_with_no_action_and_old_sequence_data(
    base_store: Store, helpers: StoreHelpers, timestamp_yesterday: datetime
):
    """Test that a case is not returned when cases with no action, but old sequence data."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(
        base_store, sequenced_at=timestamp_yesterday, is_external=True
    )

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(base_store, pipeline=Pipeline.MIP_DNA)

    # Given an action set to None
    test_analysis.family.action: Union[None, str] = None

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_analysis.family, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = get_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_get_cases_with_scout_data_delivery(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned when Scout is specified as a data delivery option."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store)

    # GIVEN a case with Scout as data delivery
    test_case = helpers.add_case(base_store, data_delivery=DataDelivery.FASTQ_ANALYSIS_SCOUT)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases with Scout as data delivery option
    cases: Query = get_cases_with_scout_data_delivery(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_get_report_supported_data_delivery_cases(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned for a delivery report supported data delivery option."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store)

    # GIVEN a case with Scout and a not supported option as data deliveries
    test_case = helpers.add_case(base_store, data_delivery=DataDelivery.FASTQ_ANALYSIS_SCOUT)
    test_invalid_case = helpers.add_case(base_store, name="test", data_delivery=DataDelivery.FASTQ)

    # GIVEN a database with the test cases
    base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.relate_sample(test_invalid_case, test_sample, PhenotypeStatus.UNKNOWN)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN retrieving the delivery report supported cases
    cases: Query = get_report_supported_data_delivery_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN only the delivery report supported case should be retrieved
    assert test_case in cases
    assert test_invalid_case not in cases


def test_get_inactive_analysis_cases(base_store: Store, helpers: StoreHelpers):
    """Test that an inactive case is returned when there is case which has no action set."""

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Family)

    # WHEN getting completed cases
    cases: Query = get_inactive_analysis_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases

    assert cases.all()[0].internal_id == test_case.internal_id


def test_get_inactive_analysis_cases_when_on_hold(base_store: Store, helpers: StoreHelpers):
    """Test that an inactivated case is returned when there is case which has action set to hold."""

    # GIVEN a case
    test_case = helpers.add_case(base_store, action=CaseActions.HOLD)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Family)

    # WHEN getting completed cases
    cases: Query = get_inactive_analysis_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases

    assert cases[0].internal_id == test_case.internal_id


def test_get_inactive_analysis_cases_when_not_completed(base_store: Store, helpers: StoreHelpers):
    """Test that no case is returned when there is case which action set to running."""

    # GIVEN a case
    helpers.add_case(base_store, action=CaseActions.RUNNING)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Family)

    # WHEN getting completed cases
    cases: Query = get_inactive_analysis_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_get_old_cases(base_store: Store, helpers: StoreHelpers, timestamp_in_2_weeks: datetime):
    """Test that an old case is returned when a future date is supplied."""

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Family)

    # WHEN getting completed cases
    cases: Query = get_older_cases_by_creation_date(cases=cases, creation_date=timestamp_in_2_weeks)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases

    assert cases.all()[0].internal_id == test_case.internal_id


def test_get_old_cases_none_when_all_cases_are_too_new(
    base_store: Store, helpers: StoreHelpers, timestamp_yesterday: datetime
):
    """No cases are returned when all cases in the store are too new."""

    # GIVEN a case
    helpers.add_case(base_store)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Family)

    # WHEN getting completed cases
    cases: Query = get_older_cases_by_creation_date(cases=cases, creation_date=timestamp_yesterday)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_get_newer_cases_none_when_all_cases_older(
    base_store: Store, helpers: StoreHelpers, timestamp_in_2_weeks: datetime
):
    """Test that no cases are returned when all cases are older than the given date."""

    # GIVEN a cases Query
    helpers.add_case(base_store)
    cases: Query = base_store._get_query(table=Family)

    # WHEN getting newer cases
    cases: Query = get_newer_cases_by_creation_date(cases=cases, creation_date=timestamp_in_2_weeks)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN no cases should be returned
    assert not cases.all()


def test_get_newer_cases_all_when_all_cases_newer(
    base_store: Store, helpers: StoreHelpers, timestamp_yesterday: datetime
):
    """Test that all cases are returned when all cases are newer than the given date."""
    # GIVEN a cases Query
    helpers.add_case(base_store)
    cases: Query = base_store._get_query(table=Family)

    # WHEN getting newer cases
    filtered_cases: Query = get_newer_cases_by_creation_date(
        cases=cases, creation_date=timestamp_yesterday
    )

    # ASSERT that cases is a query
    assert isinstance(filtered_cases, Query)

    # THEN all cases should be returned
    assert filtered_cases.count() == cases.count()


def test_filter_case_by_existing_entry_id(store_with_multiple_cases_and_samples: Store):
    # GIVEN a store containing a case with an entry id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    entry_id: int = cases_query.first().id
    assert entry_id

    # WHEN filtering for cases with the entry_id
    cases: Query = filter_cases_by_entry_id(cases=cases_query, entry_id=entry_id)

    # THEN the case should have the entry_id
    assert cases.first().id == entry_id


def test_filter_cases_by_non_existing_entry_id(
    store_with_multiple_cases_and_samples: Store, non_existent_id: str
):
    # GIVEN a store containing cases without a specific entry id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    entry_ids = [case.id for case in cases_query.all()]
    assert non_existent_id not in entry_ids

    # WHEN filtering for cases with the non existing entry id
    cases: Query = filter_cases_by_entry_id(cases=cases_query, entry_id=non_existent_id)

    # THEN the query should contain no cases
    assert cases.count() == 0


def test_filter_case_by_existing_internal_id(
    store_with_multiple_cases_and_samples: Store, case_id: str
):
    # GIVEN a store containing a case with an internal id case_id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    internal_ids = [case.internal_id for case in cases_query.all()]
    assert case_id in internal_ids

    # WHEN filtering for cases with the internal id case_id
    cases: Query = filter_case_by_internal_id(cases=cases_query, internal_id=case_id)

    # THEN the query should contain only one case
    assert cases.count() == 1

    # THEN the case should have the internal id case_id
    assert cases.first().internal_id == case_id


def test_filter_cases_by_non_existing_internal_id(
    store_with_multiple_cases_and_samples: Store, non_existent_id: str
):
    # GIVEN a store containing a case with an internal id case_id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    internal_ids = [case.internal_id for case in cases_query.all()]
    assert non_existent_id not in internal_ids

    # WHEN filtering for cases with the internal id case_id
    cases: Query = filter_case_by_internal_id(cases=cases_query, internal_id=non_existent_id)

    # THEN the query should contain no cases
    assert cases.count() == 0


def test_filter_case_by_empty_internal_id(store_with_multiple_cases_and_samples: Store):
    # GIVEN a store containing cases
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)

    # WHEN filtering for cases with an empty internal id
    cases: Query = filter_case_by_internal_id(cases=cases_query, internal_id="")

    # THEN the query should return no cases
    assert cases.count() == 0


def test_get_active_cases_no_running_cases(store_with_multiple_cases_and_samples: Store):
    """Test that no cases are returned when no cases have a running action."""
    # GIVEN a store containing cases with no "running" action
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    cases_query = cases_query.filter(Family.action != "running")

    # WHEN getting active cases
    active_cases: Query = get_running_cases(cases=cases_query)

    # THEN the query should return no cases
    assert active_cases.count() == 0


def test_get_active_cases_with_running_cases(store_with_multiple_cases_and_samples: Store):
    """Test that at least one case is returned when at least one case has a running action."""
    # GIVEN a store containing cases with at least one "running" action
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    actions: List[str] = [case.action for case in cases_query.all()]
    assert "running" in actions

    # WHEN getting active cases
    active_cases: Query = get_running_cases(cases=cases_query)

    # THEN the query should return at least one case
    assert active_cases.count() >= 1


def test_get_active_cases_only_running_cases(store_with_multiple_cases_and_samples: Store):
    """Test that all cases are returned when all cases have a running action."""
    # GIVEN a store containing only cases with "running" action
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    for case in cases_query.all():
        case.action = "running"

    # WHEN getting active cases
    active_cases: Query = get_running_cases(cases=cases_query)

    # THEN the query should return the same number of cases as the original query
    assert active_cases.count() == cases_query.count()


def test_filter_cases_by_ticket_no_matching_ticket(
    store_with_multiple_cases_and_samples: Store, non_existent_id: str
):
    """Test that no cases are returned when filtering by a non-existent ticket."""
    # GIVEN a store containing cases with no matching ticket id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)

    # WHEN filtering cases by a non-existent ticket
    filtered_cases: Query = filter_cases_by_ticket_id(cases=cases_query, ticket_id=non_existent_id)

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_ticket_matching_ticket(
    store_with_multiple_cases_and_samples: Store, ticket_id: str
):
    """Test that cases are returned when filtering by an existing ticket id."""
    # GIVEN a store containing cases with a matching ticket id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)

    # WHEN filtering cases by an existing ticket id
    filtered_cases: Query = filter_cases_by_ticket_id(cases=cases_query, ticket_id=ticket_id)

    # THEN the query should return cases with the matching ticket
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert ticket_id in case.tickets


def test_filter_cases_by_customer_entry_ids(store_with_multiple_cases_and_samples: Store):
    """Test that cases are returned when filtering by customer entry ids."""
    # GIVEN a store containing cases with customer ids
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    customer_ids = [case.customer_id for case in cases_query.all()]
    assert customer_ids

    # WHEN filtering cases by customer ids
    filtered_cases: Query = filter_cases_by_customer_entry_ids(
        cases=cases_query, customer_entry_ids=customer_ids
    )

    # THEN the filtered_cases should have the same count as cases_query
    assert filtered_cases.count() == cases_query.count()

    # THEN all cases in filtered_cases should have a customer_id in the customer_ids list
    for case in filtered_cases:
        assert case.customer_id in customer_ids


def test_filter_cases_by_name(store_with_multiple_cases_and_samples: Store):
    """Test that cases are returned when filtering by name."""
    # GIVEN a store containing cases with various names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    test_name = cases_query.first().name

    # WHEN filtering cases by a specific name
    filtered_cases: Query = filter_cases_by_name(cases=cases_query, name=test_name)

    # THEN all cases in filtered_cases should have the specified name
    for case in filtered_cases:
        assert case.name == test_name


def test_filter_cases_by_search_pattern(store_with_multiple_cases_and_samples: Store):
    """Test that cases are returned when filtering by matching internal ids."""
    # GIVEN a store containing cases with internal ids and names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    test_internal_id_pattern = cases_query.first().internal_id[:3]
    test_name_pattern = cases_query.first().name[:3]

    # WHEN filtering cases by matching internal id or name
    filtered_cases: Query = filter_cases_by_case_search(
        cases=cases_query,
        case_search=test_internal_id_pattern,
        name_search_pattern=test_name_pattern,
    )

    # THEN at least one case in filtered_cases should have an internal_id or name matching the specified patterns
    assert any(
        case.internal_id.startswith(test_internal_id_pattern)
        or case.name.startswith(test_name_pattern)
        for case in filtered_cases
    )


def test_filter_cases_not_analysed_no_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when all cases have been analysed."""
    # GIVEN a store containing only cases that have been analysed
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    for case in cases_query.all():
        case.analyses.append(Analysis(completed_at=datetime.now()))

    # WHEN filtering cases for cases that have not been analysed
    filtered_cases: Query = filter_cases_not_analysed(cases=cases_query)

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_not_analysed_no_cases_in_progress(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that all cases are returned when no cases are in progress."""
    # GIVEN a store containing cases that have not been analysed and no cases are in progress
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    for case in cases_query.all():
        case.analyses = []
        case.action = CaseActions.HOLD

    # WHEN filtering cases for cases that have not been analysed
    filtered_cases: Query = filter_cases_not_analysed(cases=cases_query)

    # THEN the query should return all cases
    assert filtered_cases.count() == cases_query.count()


def test_filter_cases_not_analysed_in_progress(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases in progress are returned."""
    # GIVEN a store containing cases that have not been analysed and at least one case is in progress
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    for case in cases_query.all():
        case.analyses = []
        case.action = CaseActions.ANALYZE

    # WHEN filtering cases for cases that have not been analysed
    filtered_cases: Query = filter_cases_not_analysed(cases=cases_query)

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_pipeline_search_no_matching_pipeline(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with matching pipeline search."""
    # GIVEN a store containing cases with different pipeline names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    pipeline_search = "non_existent_pipeline"

    # WHEN filtering cases by a non-matching pipeline search
    filtered_cases: Query = filter_cases_by_pipeline_search(
        cases=cases_query, pipeline_search=pipeline_search
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_pipeline_search_partial_match(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with partially matching pipeline search are returned."""
    # GIVEN a store containing cases with different pipeline names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    pipeline_search = cases_query.first().data_analysis[:3]

    # WHEN filtering cases by a partially matching pipeline search
    filtered_cases: Query = filter_cases_by_pipeline_search(
        cases=cases_query, pipeline_search=pipeline_search
    )

    # THEN the query should return the cases with partially matching pipeline names
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert pipeline_search in case.data_analysis


def test_filter_cases_by_pipeline_search_exact_match(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with exactly matching pipeline search are returned."""
    # GIVEN a store containing cases with different pipeline names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    pipeline_search = cases_query.first().data_analysis

    # WHEN filtering cases by an exactly matching pipeline search
    filtered_cases: Query = filter_cases_by_pipeline_search(
        cases=cases_query, pipeline_search=pipeline_search
    )

    # THEN the query should return the cases with exactly matching pipeline names
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.data_analysis == pipeline_search


def test_filter_cases_by_priority_no_matching_priority(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with matching priority."""
    # GIVEN a store containing cases with different priorities
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    non_existent_priority = "non_existent_priority"

    # WHEN filtering cases by a non-matching priority
    filtered_cases: Query = filter_cases_by_priority(
        cases=cases_query, priority=non_existent_priority
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_priority_matching_priority(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with matching priority are returned."""
    # GIVEN a store containing cases with different priorities
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    existing_priority = cases_query.first().priority

    # WHEN filtering cases by a matching priority
    filtered_cases: Query = filter_cases_by_priority(cases=cases_query, priority=existing_priority)

    # THEN the query should return the cases with matching priority
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.priority == existing_priority


def test_filter_cases_by_priority_all_priorities(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that filtering cases by all available priorities returns all cases."""
    # GIVEN a store containing cases with different priorities
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    all_priorities = set(case.priority for case in cases_query)

    # WHEN filtering cases by all available priorities
    filtered_cases = []
    for priority in all_priorities:
        filtered_cases.extend(filter_cases_by_priority(cases=cases_query, priority=priority).all())

    # THEN the query should return all cases
    assert len(filtered_cases) == cases_query.count()


def test_get_newer_cases_by_order_date_no_newer_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with a newer order date."""
    # GIVEN a store containing cases with different order dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    latest_order_date = max(case.ordered_at for case in cases_query)

    # WHEN filtering cases by a date that is later than the latest order date
    filtered_cases: Query = get_newer_cases_by_order_date(
        cases=cases_query, order_date=latest_order_date
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_get_newer_cases_by_order_date_some_newer_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with order dates newer than the given date are returned."""
    # GIVEN a store containing cases with different order dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    min_order_date = min(case.ordered_at for case in cases_query)
    max_order_date = max(case.ordered_at for case in cases_query)

    # Calculate an intermediate date between the minimum and maximum order dates
    some_order_date = min_order_date + (max_order_date - min_order_date) / 2

    # WHEN filtering cases by a date that is earlier than some order dates
    filtered_cases: Query = get_newer_cases_by_order_date(
        cases=cases_query, order_date=some_order_date
    )

    # THEN the query should return the cases with order dates newer than the given date
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.ordered_at > some_order_date


def test_get_older_cases_by_created_date_no_older_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with an older creation date."""
    # GIVEN a store containing cases with different creation dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    oldest_created_date = min(case.created_at for case in cases_query)

    # WHEN filtering cases by a date that is later than the oldest order date
    filtered_cases: Query = get_older_cases_by_creation_date(
        cases=cases_query, creation_date=oldest_created_date
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_get_older_cases_by_creation_date_some_newer_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with creation dates older than the given date are returned."""
    # GIVEN a store containing cases with different creation dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)
    min_creation_date = min(case.created_at for case in cases_query)
    max_creation_date = max(case.created_at for case in cases_query)

    # Calculate an intermediate date between the minimum and maximum creation dates
    some_creation_date = min_creation_date + (max_creation_date - min_creation_date) / 2

    # WHEN filtering cases by a date that is earlier than some creation dates
    filtered_cases: Query = get_older_cases_by_creation_date(
        cases=cases_query, creation_date=some_creation_date
    )

    # THEN the query should return the cases with creation dates older than the given date
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.created_at < some_creation_date



"""Tests for BALSAMIC analysis"""
from pathlib import Path

import pytest
from _pytest.logging import LogCaptureFixture

from cg.constants.constants import SampleType
from cg.constants.observations import ObservationsFileWildcards
from cg.constants.sequencing import Variants
from cg.constants.subject import Gender
from cg.exc import BalsamicStartError

from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig


def test_get_verified_gender():
    """Tests gender extraction from a sample dictionary"""

    # GIVEN a sample object
    sample_obj = {
        "ACC0000A0": {"gender": "female"},
        "ACC0000A1": {"gender": "female"},
    }

    # WHEN extracting the gender
    retrieved_gender: Gender = BalsamicAnalysisAPI.get_verified_gender(sample_obj)

    # THEN gender must match the expected one
    assert retrieved_gender == "female"


def test_get_verified_gender_error():
    """Tests gender extraction from a sample dictionary when two different gender has been provided"""

    # GIVEN a sample object with different genders
    sample_obj = {
        "ACC0000A0": {"gender": "male"},
        "ACC0000A1": {"gender": "female"},
    }

    # WHEN extracting the gender
    with pytest.raises(BalsamicStartError):
        BalsamicAnalysisAPI.get_verified_gender(sample_obj)


def test_get_verified_gender_unknown(caplog):
    """Tests gender extraction from a sample dictionary when the gender is unknown"""

    # GIVEN a sample object with different genders
    sample_obj = {
        "ACC0000A0": {"gender": "unknown"},
        "ACC0000A1": {"gender": "unknown"},
    }

    # WHEN extracting the gender
    retrieved_gender: Gender = BalsamicAnalysisAPI.get_verified_gender(sample_obj)

    # THEN gender must match the expected one
    assert retrieved_gender == Gender.FEMALE
    assert f"The provided gender is unknown, setting {Gender.FEMALE} as the default" in caplog.text


def test_get_verified_pon():
    """Tests PON verification"""

    # GIVEN specific panel bed and PON files
    panel_bed = "/path/gmcksolid_4.1_hg19_design.bed"
    pon_cnn = "/path/PON/gmcksolid_4.1_hg19_design_CNVkit_PON_reference_v2.cnn"
    invalid_pon_cnn = "/path/PON/gmssolid_15.2_hg19_design_CNVkit_PON_reference_v2.cnn"

    # WHEN validating the PON
    validated_pon: str = BalsamicAnalysisAPI.get_verified_pon(None, panel_bed, pon_cnn)

    # THEN the PON verification should be performed successfully
    assert pon_cnn == validated_pon
    with pytest.raises(BalsamicStartError):
        BalsamicAnalysisAPI.get_verified_pon(None, panel_bed, invalid_pon_cnn)


def test_get_latest_file_by_pattern(
    cg_context: CGConfig, observations_dir: Path, observations_somatic_snv_file_path: Path
):
    """Test latest observations extraction."""

    # GIVEN a Loqusdb temporary directory and a cancer SNV file wildcard
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)

    # WHEN getting the latest observations file
    observation: str = balsamic_analysis_api.get_latest_file_by_pattern(
        directory=observations_dir, pattern=ObservationsFileWildcards.CANCER_SOMATIC_SNV
    )

    # THEN the extracted observation should match the latest file
    assert observation == observations_somatic_snv_file_path.as_posix()


def test_get_parsed_observation_file_paths_no_args(
    cg_context: CGConfig,
    observations_dir: Path,
    observations_clinical_sv_file_path: Path,
    observations_somatic_snv_file_path: Path,
    outdated_observations_somatic_snv_file_path: Path,
):
    """Test verified observations extraction with no arguments."""

    # GIVEN a Loqusdb temporary directory
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)
    balsamic_analysis_api.loqusdb_path = observations_dir

    # WHEN getting the latest observations arguments dictionary
    args: dict = balsamic_analysis_api.get_parsed_observation_file_paths(None)

    # THEN only the created observations files should be returned
    assert args[ObservationsFileWildcards.CLINICAL_SNV] is None
    assert (
        args[ObservationsFileWildcards.CLINICAL_SV] == observations_clinical_sv_file_path.as_posix()
    )
    assert args[ObservationsFileWildcards.CANCER_ALL_SNV] is None
    assert (
        args[ObservationsFileWildcards.CANCER_SOMATIC_SNV]
        == observations_somatic_snv_file_path.as_posix()
    )
    assert (
        outdated_observations_somatic_snv_file_path.as_posix()
        not in args[ObservationsFileWildcards.CANCER_SOMATIC_SNV]
    )
    assert args[ObservationsFileWildcards.CANCER_SOMATIC_SV] is None


def test_get_parsed_observation_file_paths_overwrite_input(
    cg_context: CGConfig,
    observations_dir: Path,
    observations_clinical_snv_file_path: Path,
    custom_observations_clinical_snv_file_path: Path,
):
    """Test verified observations extraction when providing a non default observation file."""

    # GIVEN a Loqusdb temporary directory and a custom Loqusdb file
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)
    balsamic_analysis_api.loqusdb_path = observations_dir

    # WHEN getting the latest observations dictionary
    args: dict = balsamic_analysis_api.get_parsed_observation_file_paths(
        [custom_observations_clinical_snv_file_path.as_posix()]
    )

    # THEN the default file should be overwritten by the custom file
    assert (
        observations_clinical_snv_file_path.as_posix()
        not in args[ObservationsFileWildcards.CLINICAL_SNV]
    )
    assert (
        args[ObservationsFileWildcards.CLINICAL_SNV]
        == custom_observations_clinical_snv_file_path.as_posix()
    )


def test_get_swegen_verified_path(
    cg_context: CGConfig, swegen_dir: Path, swegen_snv_reference: Path
):
    """Test verified SweGen path return."""

    # GIVEN A SNV variants pattern and a BALSAMIC analysis API
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)
    balsamic_analysis_api.swegen_path = swegen_dir

    # WHEN obtaining the latest file by pattern
    swegen_file: str = balsamic_analysis_api.get_swegen_verified_path(Variants.SNV)

    # THEN the returned file should match the SweGen SNV reference
    assert swegen_file == swegen_snv_reference.as_posix()



from click.testing import CliRunner

from cg.cli.workflow.commands import (
    balsamic_past_run_dirs,
    balsamic_qc_past_run_dirs,
    balsamic_umi_past_run_dirs,
    balsamic_pon_past_run_dirs,
    fluffy_past_run_dirs,
    microsalt_past_run_dirs,
    mip_dna_past_run_dirs,
    mip_rna_past_run_dirs,
    mutant_past_run_dirs,
    rnafusion_past_run_dirs,
)
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_cli_workflow_clean_balsamic(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    """Test clean Balsamic workflow."""
    # GIVEN a before string

    # WHEN running command
    result = cli_runner.invoke(balsamic_past_run_dirs, [before_date], obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_balsamic_qc(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    # GIVEN a before string

    # WHEN running command in dry-run
    result = cli_runner.invoke(balsamic_qc_past_run_dirs, [before_date], obj=base_context)

    # THEN command should terminate successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_balsamic_pon(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    # GIVEN a before string

    # WHEN running command in dry-run
    result = cli_runner.invoke(balsamic_pon_past_run_dirs, [before_date], obj=base_context)

    # THEN command should terminate successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_balsamic_umi(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    # GIVEN a before string

    # WHEN running command in dry-run
    result = cli_runner.invoke(balsamic_umi_past_run_dirs, [before_date], obj=base_context)

    # THEN command should terminate successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_fluffy(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    """Test clean Fluffy workflow."""
    # GIVEN a before string

    # WHEN running command
    result = cli_runner.invoke(fluffy_past_run_dirs, [before_date], obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_mip_dna(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    """Test clean MIP DNA workflow."""
    # GIVEN a before string

    # WHEN running command
    result = cli_runner.invoke(mip_dna_past_run_dirs, [before_date], obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_mip_rna(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    """Test clean MIP RNA workflow."""
    # GIVEN a before string

    # WHEN running command
    result = cli_runner.invoke(mip_rna_past_run_dirs, [before_date], obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_mutant(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    """Test clean Mutan workflow."""
    # GIVEN a before string

    # WHEN running command
    result = cli_runner.invoke(mutant_past_run_dirs, [before_date], obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_rnafusion(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    """Test clean RNAFusion workflow."""
    # GIVEN a before string

    # WHEN running command
    result = cli_runner.invoke(rnafusion_past_run_dirs, [before_date], obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_microsalt(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):
    """Test clean Micrsalt workflow."""
    # Given a before string

    # WHEN running command
    result = cli_runner.invoke(microsalt_past_run_dirs, [before_date], obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS



"""Tests for RNA part of the scout upload API"""
import logging
from typing import Generator, List
import pytest
from _pytest.logging import LogCaptureFixture

from sqlalchemy.orm import Query
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Pipeline
from cg.constants.sequencing import SequencingMethod
from cg.exc import CgDataError
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.store.models import Family, Sample
import cg.store as Store
from tests.store_helpers import StoreHelpers


def set_is_tumour_on_case(store: Store, case_id: str, is_tumour: bool):
    for link in store.get_case_by_internal_id(internal_id=case_id).links:
        link.sample.is_tumour = is_tumour


def get_subject_id_from_case(store: Store, case_id: str) -> str:
    for link in store.get_case_by_internal_id(internal_id=case_id).links:
        return link.sample.subject_id


def ensure_two_dna_tumour_matches(
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    rna_case_id: str,
    rna_store: Store,
) -> None:
    """Ensures that we have one RNA case that has two matching DNA cases via subject id and tumour state."""
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    subject_id: str = get_subject_id_from_case(store=rna_store, case_id=rna_case_id)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=True)
    dna_extra_case = helpers.ensure_case(
        store=rna_store, customer=rna_store.get_case_by_internal_id(dna_case_id).customer
    )
    another_sample_id = helpers.add_sample(
        store=rna_store,
        name=another_sample_id,
        subject_id=subject_id,
        is_tumour=True,
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    helpers.add_relationship(store=rna_store, sample=another_sample_id, case=dna_extra_case)
    rna_store.session.commit()


def ensure_extra_rna_case_match(
    another_rna_sample_id: str,
    helpers: StoreHelpers,
    rna_case_id: str,
    rna_store: Store,
) -> None:
    """Ensures that we have an extra RNA case that matches by subject_id the existing RNA case and DNA cases."""
    rna_extra_case = helpers.ensure_case(
        store=rna_store,
        data_analysis=Pipeline.MIP_RNA,
        customer=rna_store.get_case_by_internal_id(rna_case_id).customer,
    )
    subject_id: str = get_subject_id_from_case(store=rna_store, case_id=rna_case_id)
    another_rna_sample_id = helpers.add_sample(
        store=rna_store,
        internal_id=another_rna_sample_id,
        subject_id=subject_id,
        is_tumour=False,
        application_type=SequencingMethod.WTS,
    )
    helpers.add_relationship(store=rna_store, sample=another_rna_sample_id, case=rna_extra_case)


def test_upload_rna_junctions_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper
    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_rna_junctions_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the 2 files should have been uploaded to the connected sample on the dna case in scout
    assert "Upload splice junctions bed file finished!" in caplog.text
    assert "Upload RNA coverage bigwig file finished!" in caplog.text


def test_upload_splice_junctions_bed_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the splice junctions file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload splice junctions bed file finished!" in caplog.text

    # THEN the customers dna samples name should have been mentioned in the logging (and used in the upload)
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name in caplog.text


def test_upload_rna_coverage_bigwig_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's bigWig file for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the bigWig file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload RNA coverage bigwig file finished!" in caplog.text

    # THEN the customers dna samples name should have been mentioned in the logging (and used in the upload)
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name in caplog.text


def test_upload_clinical_rna_fusion_report_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's clinical fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a clinical fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the fusion report file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload Clinical fusion report finished!" in caplog.text

    # THEN the dna case id should have been mentioned in the logging (and used in the upload)
    assert dna_case_id in caplog.text

    # THEN the customers dna samples name should NOT have been mentioned in the logging
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name not in caplog.text


def test_upload_research_rna_fusion_report_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True, research=True)

    # THEN the fusion report file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload Research fusion report finished!" in caplog.text

    # THEN the dna case id should have been mentioned in the logging (and used in the upload)
    assert dna_case_id in caplog.text

    # THEN the customers dna samples name should NOT have been mentioned in the logging
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name not in caplog.text


def test_upload_rna_fusion_report_to_scout_no_subject_id(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    for link in rna_store.get_case_by_internal_id(rna_case_id).links:
        link.sample.subject_id = ""
    for link in rna_store.get_case_by_internal_id(dna_case_id).links:
        link.sample.subject_id = ""
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_coverage_bigwig_to_scout_no_subject_id(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    for link in rna_store.get_case_by_internal_id(rna_case_id).links:
        link.sample.subject_id = ""
    for link in rna_store.get_case_by_internal_id(dna_case_id).links:
        link.sample.subject_id = ""
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_splice_junctions_bed_to_scout_no_subject_id(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    for link in rna_store.get_case_by_internal_id(rna_case_id).links:
        link.sample.subject_id = ""
    for link in rna_store.get_case_by_internal_id(dna_case_id).links:
        link.sample.subject_id = ""
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_fusion_report_to_scout_tumour_non_matching(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that an RNA case's gene fusion report is not uploaded if the is_tumour is not matching"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via is_tumour (i.e. different is_tumour)
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=False)
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_coverage_bigwig_to_scout_tumour_non_matching(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples is not uploaded if the is_tumour is not matching"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via is_tumour (i.e. different is_tumour)
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=False)
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_splice_junctions_bed_to_scout_tumour_non_matching(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples is not uploaded if the is_tumour is not matching"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via is_tumour (i.e. different is_tumour)
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=False)
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_fusion_report_to_scout_tumour_multiple_matches(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that an RNA case's gene fusion report is not uploaded if the is_tumour has too many DNA-matches"""

    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via is_tumour (i.e. same is_tumour)
    ensure_two_dna_tumour_matches(dna_case_id, another_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_coverage_bigwig_to_scout_tumour_multiple_matches(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples is not uploaded if the RNA-sample has too many DNA-matches"""

    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via is_tumour (i.e. same is_tumour)
    ensure_two_dna_tumour_matches(dna_case_id, another_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_splice_junctions_bed_to_scout_tumour_multiple_matches(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples is not uploaded if the RNA-sample has too many DNA-matches"""

    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via is_tumour (i.e. same is_tumour)
    ensure_two_dna_tumour_matches(dna_case_id, another_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)


def test_get_application_prep_category(
    another_rna_sample_id: str,
    dna_sample_son_id: str,
    helpers: StoreHelpers,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that RNA samples are removed when filtering sample list by pipeline"""

    # GIVEN an RNA sample that is connected by subject ID to one RNA and one DNA sample in other cases

    ensure_extra_rna_case_match(another_rna_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    dna_sample: Sample = rna_store.get_sample_by_internal_id(dna_sample_son_id)
    another_rna_sample_id: Sample = rna_store.get_sample_by_internal_id(another_rna_sample_id)
    all_son_rna_dna_samples: List[Sample] = [dna_sample, another_rna_sample_id]

    # WHEN running the method to filter a list of Sample objects containing RNA and DNA samples connected by subject_id
    only_son_dna_samples = upload_scout_api._get_application_prep_category(all_son_rna_dna_samples)

    # THEN even though an RNA sample is present in the initial query, the output should not contain any RNA samples
    nr_of_subject_id_samples: int = len(all_son_rna_dna_samples)
    nr_of_subject_id_dna_samples: int = len([only_son_dna_samples])
    assert nr_of_subject_id_samples == 2
    assert nr_of_subject_id_dna_samples == 1


def test_create_rna_dna_sample_case_map(
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that the create_rna_dna_sample_case_map returns a nested dictionary."""

    # GIVEN an RNA case with RNA samples that are connected by subject ID to DNA samples in a DNA case
    rna_case: Family = rna_store.get_case_by_internal_id(internal_id=rna_case_id)

    # WHEN running the method to create a nested dictionary with the relationships between RNA/DNA samples and DNA cases
    rna_dna_case_map: dict = upload_scout_api.create_rna_dna_sample_case_map(rna_case=rna_case)

    # THEN the output should be a nested dictionary for each key: {key:{value:[]}}
    assert all(isinstance(items, dict) for items in rna_dna_case_map.values())


def test_add_rna_sample(
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that for a given RNA case the RNA samples are added to the rna_dna_case_map."""

    # GIVEN an RNA case and the associated RNA samples
    rna_case: Family = rna_store.get_case_by_internal_id(internal_id=rna_case_id)
    rna_sample_list: List[Sample] = rna_store.get_samples_by_name_pattern(name_pattern="rna")

    # WHEN running the method to create a nested dictionary with the relationships between RNA/DNA samples and DNA cases
    rna_dna_case_map: dict = upload_scout_api.create_rna_dna_sample_case_map(rna_case=rna_case)

    # THEN the resulting dictionary should contain all RNA samples in the case
    for key in rna_sample_list:
        assert key.internal_id in list(rna_dna_case_map.keys())


def test_link_rna_sample_to_dna_sample(
    dna_sample_son_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test for a given RNA sample, the associated DNA sample name matches and is present in rna_dna_case_map."""

    # GIVEN an RNA sample
    rna_sample: Sample = rna_store.get_sample_by_internal_id(rna_sample_son_id)

    # WHEN adding the RNA sample to the rna_dna_case_map
    rna_dna_case_map: dict = {}
    upload_scout_api._add_rna_sample(
        rna_sample=rna_sample, rna_dna_sample_case_map=rna_dna_case_map
    )

    # THEN the rna_dna_case_map should contain the RNA sample
    assert rna_sample_son_id in rna_dna_case_map

    # THEN the rna_dna_case_map values should contain the associated DNA sample
    dna_samples: dict = rna_dna_case_map[rna_sample.internal_id]
    assert dna_sample_son_id in dna_samples


def test_add_dna_cases_to_dna_sample(
    dna_case_id: str,
    dna_sample_son_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test for a given RNA sample, the DNA case name matches to the case name of the DNA sample in rna_dna_case_map."""

    # GIVEN an RNA sample, a DNA sample, and a DNA case
    rna_sample: Sample = rna_store.get_sample_by_internal_id(rna_sample_son_id)
    dna_sample: Sample = rna_store.get_sample_by_internal_id(dna_sample_son_id)
    dna_case: Family = rna_store.get_case_by_internal_id(internal_id=dna_case_id)

    # WHEN adding the RNA sample rna_dna_case_map
    rna_dna_case_map: dict = {}
    upload_scout_api._add_rna_sample(
        rna_sample=rna_sample, rna_dna_sample_case_map=rna_dna_case_map
    )

    # THEN the rna_dna_case_map should contain the DNA_case name associated with the DNA sample
    case_names: list = rna_dna_case_map[rna_sample.internal_id][dna_sample.name]
    assert dna_case.internal_id in case_names



"""Base command for trending"""

import datetime as dt
import logging
from pathlib import Path
from typing import Any, List, Optional, Tuple, Dict

import click

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Pipeline
from cg.exc import AnalysisUploadError
from cg.meta.upload.vogue import UploadVogueAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import FamilySample, Analysis
from housekeeper.store.models import File, Version
from cg.utils.dispatcher import Dispatcher

LOG = logging.getLogger(__name__)

VOGUE_VALID_BIOINFO = [str(Pipeline.BALSAMIC)]


def validate_date(context, param, date_: str) -> Optional[dt.date]:
    """validates the cli input of a date (if any) and returns the date in the proper format"""
    if not date_:
        return
    try:
        return dt.datetime.strptime(date_, "%Y-%m-%d").date() or None
    except ValueError:
        raise click.BadParameter("Date should be in ISO format: YYYY-MM-DD")


@click.group()
@click.pass_obj
def vogue(context: CGConfig):
    """Load trending data into trending database"""

    if not context.meta_apis.get("analysis_api"):
        context.meta_apis["analysis_api"] = MipDNAAnalysisAPI(context)

    click.echo(click.style("----------------- TRENDING -----------------------"))


@vogue.command("genotype", short_help="Getting genotype data from the genotype database.")
@click.option(
    "-d", "--days", type=int, required="True", help="load X days old sampels from genotype to vogue"
)
@click.pass_obj
def genotype(context: CGConfig, days: int):
    """Loading samples from the genotype database to the trending database"""

    upload_vogue_api = UploadVogueAPI(
        genotype_api=context.genotype_api, vogue_api=context.vogue_api, store=context.status_db
    )
    click.echo(click.style("----------------- GENOTYPE -----------------------"))

    upload_vogue_api.load_genotype(days=days)


@vogue.command("apptags", short_help="Getting application tags to the trending database.")
@click.pass_obj
def apptags(context: CGConfig):
    """Loading apptags from status db to the trending database"""

    upload_vogue_api = UploadVogueAPI(
        genotype_api=context.genotype_api, vogue_api=context.vogue_api, store=context.status_db
    )
    click.echo(click.style("----------------- APPLICATION TAGS -----------------------"))
    upload_vogue_api.load_apptags()


@vogue.command("flowcells", short_help="Getting flowcell data from the lims.")
@click.option(
    "-d", "--days", type=int, required="True", help="load X days old runs from lims to vogue"
)
@click.pass_obj
def flowcells(context: CGConfig, days: int):
    """Loading runs from lims to the trending database"""
    LOG.info("----------------- FLOWCELLS -----------------------")
    context.vogue_api.load_flowcells(days=days)


@vogue.command("samples", short_help="Getting sample data from lims.")
@click.option(
    "-d", "--days", type=int, required="True", help="load X days old sampels from lims to vogue"
)
@click.pass_obj
def samples(context: CGConfig, days: int):
    """Loading samples from lims to the trending database"""
    LOG.info("----------------- SAMPLES -----------------------")
    context.vogue_api.load_samples(days=days)


@vogue.command("reagent-labels", short_help="Getting reagent_label data from lims.")
@click.option(
    "-d", "--days", type=int, required=True, help="load X days old sampels from lims to vogue"
)
@click.pass_obj
def reagent_labels(context: CGConfig, days: int):
    """Loading reagent_labels from lims to the trending database"""
    LOG.info("----------------- REAGENT LABELS -----------------------")
    context.vogue_api.load_reagent_labels(days=days)


@vogue.command("bioinfo", short_help="Load bioinfo results into vogue")
@click.option(
    "-c",
    "--case-name",
    required=True,
    help="Case name or project name for which the analysis results will load",
)
@click.option(
    "--cleanup/--no-cleanup", default=False, help="Cleanup processed case data while loading"
)
@click.option(
    "-t",
    "--target-load",
    default="all",
    type=click.Choice(["all", "raw", "process"]),
    show_default=True,
    help=(
        "Final target to load bioinfo data."
        "Target orders are: all, raw, process."
        "Selecting all, will load raw, process, and sample."
        "Selecting process, will upload raw and process."
        "Selecting raw, will only load raw data."
    ),
)
@click.option("--dry/--no-dry", default=False, help="Dry run...")
@click.pass_obj
def bioinfo(context: CGConfig, case_name: str, cleanup: bool, target_load: str, dry: bool):
    """Load bioinfo case results to the trending database"""
    status_db: Store = context.status_db
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    upload_vogue_api = UploadVogueAPI(
        genotype_api=context.genotype_api, vogue_api=context.vogue_api, store=status_db
    )

    click.echo(click.style("----------------- BIOINFO -----------------------"))

    load_bioinfo_raw_inputs = dict()

    # Probably get samples for a case_name through statusdb api
    load_bioinfo_raw_inputs["samples"] = _get_samples(status_db, case_name)

    # Probably get analysis result file through housekeeper ai
    load_bioinfo_raw_inputs["analysis_result_file"] = _get_multiqc_latest_file(
        housekeeper_api, case_name
    )

    # Probably get analysis_type [multiqc or microsalt or all] from cli
    # This might automated to some extend by checking if input multiqc json.
    # This tells us how the result was generated. If it is multiqc it will try to validate keys with
    # an actual model.
    load_bioinfo_raw_inputs["analysis_type"] = "multiqc"

    # case_name is the input
    load_bioinfo_raw_inputs["analysis_case_name"] = case_name

    # Get case_analysis_type from cli a free text for an entry in trending database
    load_bioinfo_raw_inputs["case_analysis_type"] = "multiqc"

    # Get workflow_name and workflow_version
    workflow_name, workflow_version = _get_analysis_workflow_details(status_db, case_name)

    if workflow_name is None:
        raise AnalysisUploadError(
            f"Case upload failed: {case_name}. Reason: Workflow name not found."
        )
    workflow_name = workflow_name.lower()

    if workflow_name not in VOGUE_VALID_BIOINFO:
        raise AnalysisUploadError(
            f"Case upload failed: {case_name}. Reason: Bad workflow name: {workflow_name}."
        )

    load_bioinfo_raw_inputs["analysis_workflow_name"] = workflow_name
    load_bioinfo_raw_inputs["analysis_workflow_version"] = workflow_version

    if dry:
        click.echo(click.style("----------------- DRY RUN -----------------------"))

    if target_load in ("raw", "all"):
        click.echo(click.style("----------------- UPLOAD UNPROCESSED -----------------------"))
        if not dry:
            upload_vogue_api.load_bioinfo_raw(load_bioinfo_raw_inputs)

    if target_load in ("process", "all"):
        click.echo(click.style("----------------- PROCESS CASE -----------------------"))
        if not dry:
            upload_vogue_api.load_bioinfo_process(load_bioinfo_raw_inputs, cleanup)
        click.echo(click.style("----------------- PROCESS SAMPLE -----------------------"))
        if not dry:
            upload_vogue_api.load_bioinfo_sample(load_bioinfo_raw_inputs)


@vogue.command("bioinfo-all", short_help="Load all bioinfo results into vogue")
@click.option("--dry/--no-dry", is_flag=True, help="Dry run...")
@click.option(
    "--completed-after",
    "-a",
    callback=validate_date,
    default=None,
    help="Only upload cases with an analysis that was completed after this date",
)
@click.option(
    "--completed-before",
    "-b",
    callback=validate_date,
    default=None,
    help="Only upload cases with an analysis that was completed before this date",
)
@click.pass_context
def bioinfo_all(
    context: click.Context,
    completed_after: Optional[str],
    completed_before: Optional[str],
    dry: bool,
):
    """Load all cases with recent analysis and a multiqc-json to the trending database."""

    status_db: Store = context.obj.status_db
    housekeeper_api: HousekeeperAPI = context.obj.housekeeper_api
    input_dict: Dict[str, Any] = {
        "completed_at_after": completed_after,
        "completed_at_before": completed_before,
    }
    function_dispatcher: Dispatcher = Dispatcher(
        functions=[
            status_db.get_analysis_for_vogue_upload_completed_after,
            status_db.get_analysis_for_vogue_upload_completed_before,
            status_db.get_analyses_for_vogue_upload,
        ],
        input_dict=input_dict,
    )
    analyses: List[Analysis] = function_dispatcher()
    if not analyses:
        LOG.info("No analyses found for upload to trending.")
        return

    for analysis in analyses:
        case_name: str = analysis.family.internal_id
        version_obj: Version = housekeeper_api.last_version(case_name)
        if not version_obj:
            continue

        # confirm multiqc.json exists
        multiqc_file_obj: List[File] = list(
            housekeeper_api.get_files(
                bundle=case_name, tags=["multiqc-json"], version=version_obj.id
            )
        )
        if not multiqc_file_obj:
            LOG.warning("No multiqc file found in Housekeeper for case %s", case_name)
            continue

        # confirm that file exists
        existing_multiqc_file: str = multiqc_file_obj[0].full_path
        if not Path(existing_multiqc_file).exists():
            LOG.warning("The file %s does not exist for case %s", existing_multiqc_file, case_name)
            continue

        LOG.info("Found multiqc for %s, %s", case_name, existing_multiqc_file)
        try:
            context.invoke(bioinfo, case_name=case_name, cleanup=True, target_load="all", dry=dry)
            if not dry:
                UploadVogueAPI.update_analysis_uploaded_to_vogue_date(analysis=analysis)
                status_db.session.commit()
        except AnalysisUploadError:
            LOG.error("Case upload failed: %s", case_name, exc_info=True)


def _get_multiqc_latest_file(hk_api: HousekeeperAPI, case_name: str) -> str:
    """Get latest multiqc_data.json path for a case_name
    Args:
        case_name(str): onemite
    Returns:
        multiqc_data_path(str): /path/to/multiqc.json
    """
    version_obj = hk_api.last_version(case_name)
    multiqc_json_file = hk_api.get_files(
        bundle=case_name, tags=["multiqc-json"], version=version_obj.id
    )

    if len(list(multiqc_json_file)) == 0:
        raise FileNotFoundError(f"No multiqc.json was found in housekeeper for {case_name}")

    return multiqc_json_file[0].full_path


def _get_samples(store: Store, case_name: str) -> str:
    """Get a sample string for case_name
    Args:
        case_name(str): onemite
    Returns:
        sample_names(str): ACC12345,ACC45679
    """

    link_objs: List[FamilySample] = store.get_case_by_internal_id(internal_id=case_name).links
    sample_ids = {link_obj.sample.internal_id for link_obj in link_objs}
    return ",".join(sample_ids)


def _get_analysis_workflow_details(status_api: Store, case_name: str) -> Tuple[Any, Optional[Any]]:
    """Get lowercase workflow name for a case_name
    Args:
        case_name(str): onemite
    Returns:
        workflow_name(str): balsamic
        workflow_version(str): v3.14.15
    """
    # Workflow that generated these results
    case_obj = status_api.get_case_by_internal_id(internal_id=case_name)
    workflow_name = None
    workflow_version = None
    if case_obj.analyses:
        workflow_name = case_obj.analyses[0].pipeline
        workflow_version = case_obj.analyses[0].pipeline_version

    return workflow_name, workflow_version



from cg.constants.rnafusion import RNAFUSION_ACCEPTED_STRANDEDNESS
from cg.models.nextflow.sample import NextflowSample, validator


class RnafusionSample(NextflowSample):
    """Rnafusion sample model

    Attributes:
        strandedness: rnafusion config file attributes model
    """

    strandedness: str

    @validator("strandedness", always=True)
    def valid_value_strandedness(cls, value) -> str:
        """Verify that the strandedness value is accepted."""
        assert value in RNAFUSION_ACCEPTED_STRANDEDNESS
        return "Strandedness value not valid"



from cg.apps.hermes.hermes_api import HermesApi
from cg.cli.workflow.fluffy.base import store, store_available
from cg.constants import EXIT_SUCCESS
from cg.meta.workflow.fluffy import FluffyAnalysisAPI
from cg.models.cg_config import CGConfig
from click.testing import CliRunner


def test_cli_store_dry_no_case(
    cli_runner: CliRunner,
    fluffy_case_id_non_existing: str,
    fluffy_context: CGConfig,
    caplog,
):
    caplog.set_level("ERROR")

    # GIVEN a case_id that does not exist in database

    # WHEN running command in dry-run mode
    result = cli_runner.invoke(
        store, [fluffy_case_id_non_existing, "--dry-run"], obj=fluffy_context
    )

    # THEN command does not terminate successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN logging informs about the case_id not existing
    assert fluffy_case_id_non_existing in caplog.text
    assert "could not be found" in caplog.text


def test_cli_store(
    cli_runner: CliRunner,
    fluffy_case_id_existing: str,
    deliverables_yaml_fixture_path,
    fluffy_hermes_deliverables_response_data,
    fluffy_context: CGConfig,
    timestamp_yesterday,
    caplog,
    mocker,
):
    caplog.set_level("INFO")
    fluffy_analysis_api: FluffyAnalysisAPI = fluffy_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does exist in database

    # GIVEN that case action is "running"
    fluffy_analysis_api.set_statusdb_action(case_id=fluffy_case_id_existing, action="running")

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(FluffyAnalysisAPI, "get_deliverables_file_path")
    FluffyAnalysisAPI.get_deliverables_file_path.return_value = deliverables_yaml_fixture_path

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(FluffyAnalysisAPI, "get_date_from_file_path")
    FluffyAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = fluffy_hermes_deliverables_response_data

    # WHEN running command
    result = cli_runner.invoke(store, [fluffy_case_id_existing], obj=fluffy_context)

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN log informs that analysis was stored in Housekeeper and StatusDB
    assert "stored in Housekeeper" in caplog.text
    assert "stored in StatusDB" in caplog.text

    # THEN action of case in StatusDB is set to None
    assert not fluffy_analysis_api.status_db.get_case_by_internal_id(
        internal_id=fluffy_case_id_existing
    ).action


def test_cli_store_bundle_already_added(
    cli_runner: CliRunner,
    fluffy_case_id_existing,
    deliverables_yaml_fixture_path,
    fluffy_hermes_deliverables_response_data,
    fluffy_context: CGConfig,
    timestamp_yesterday,
    caplog,
    mocker,
):
    caplog.set_level("ERROR")
    # GIVEN a case_id that does exist in database

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(FluffyAnalysisAPI, "get_deliverables_file_path")
    FluffyAnalysisAPI.get_deliverables_file_path.return_value = deliverables_yaml_fixture_path

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(FluffyAnalysisAPI, "get_date_from_file_path")
    FluffyAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = fluffy_hermes_deliverables_response_data

    # GIVEN deliverables have already been stored in Housekeeper
    cli_runner.invoke(store, [fluffy_case_id_existing], obj=fluffy_context)

    # WHEN running command
    result = cli_runner.invoke(store, [fluffy_case_id_existing], obj=fluffy_context)

    # THEN command does not terminate successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN log informs that storing deliverables failed
    assert "Error storing deliverables" in caplog.text


def test_cli_store_available_case_is_running(
    cli_runner: CliRunner,
    fluffy_case_id_existing,
    deliverables_yaml_fixture_path,
    fluffy_hermes_deliverables_response_data,
    fluffy_context: CGConfig,
    timestamp_yesterday,
    caplog,
    mocker,
):
    caplog.set_level("INFO")
    fluffy_analysis_api: FluffyAnalysisAPI = fluffy_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does exist in database

    # GIVEN that case action is "running"
    fluffy_analysis_api.set_statusdb_action(case_id=fluffy_case_id_existing, action="running")

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(FluffyAnalysisAPI, "get_analysis_finish_path")
    FluffyAnalysisAPI.get_analysis_finish_path.return_value = deliverables_yaml_fixture_path

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(FluffyAnalysisAPI, "get_date_from_file_path")
    FluffyAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = fluffy_hermes_deliverables_response_data

    # WHEN running command
    result = cli_runner.invoke(store_available, [], obj=fluffy_context)

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN analysis data is stored in Housekeeper and StatusDB
    assert "stored in Housekeeper" in caplog.text
    assert "stored in StatusDB" in caplog.text

    # THEN log informs about eligible case
    assert fluffy_case_id_existing in caplog.text

    # THEN case action is set to None after storing
    assert not fluffy_analysis_api.status_db.get_case_by_internal_id(
        internal_id=fluffy_case_id_existing
    ).action


def test_cli_store_available_case_not_running(
    cli_runner: CliRunner,
    fluffy_case_id_existing,
    deliverables_yaml_fixture_path,
    fluffy_hermes_deliverables_response_data,
    fluffy_context: CGConfig,
    timestamp_yesterday,
    caplog,
    mocker,
):
    caplog.set_level("INFO")
    fluffy_analysis_api: FluffyAnalysisAPI = fluffy_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does exist in database

    # GIVEN that case action is None
    fluffy_analysis_api.status_db.get_case_by_internal_id(
        internal_id=fluffy_case_id_existing
    ).action = None
    fluffy_analysis_api.status_db.session.commit()

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(FluffyAnalysisAPI, "get_deliverables_file_path")
    FluffyAnalysisAPI.get_deliverables_file_path.return_value = deliverables_yaml_fixture_path

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(FluffyAnalysisAPI, "get_date_from_file_path")
    FluffyAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = fluffy_hermes_deliverables_response_data

    # WHEN running command
    result = cli_runner.invoke(store_available, [], obj=fluffy_context)

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN case is not picked up by query and is not in the log
    assert fluffy_case_id_existing not in caplog.text

    # THEN nothing is stored in Housekeeper
    assert "stored in Housekeeper" not in caplog.text



"""Commands to mimic the behaviour from cgstats.

This is a way to interact with cgstats manually. In automated production these commands will not be run.
"""
import logging
from pathlib import Path
from typing import List, Optional

import click

from cg.apps.cgstats.crud.create import create_novaseq_flowcell
from cg.apps.cgstats.stats import StatsAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.constants.cgstats import STATS_HEADER
from cg.constants.demultiplexing import OPTION_BCL_CONVERTER
from cg.exc import FlowCellError
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingNovaseqAPI
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCell

LOG = logging.getLogger(__name__)


@click.command(name="add")
@OPTION_BCL_CONVERTER
@click.argument("flow-cell-name")
@click.pass_obj
def add_flow_cell_cmd(context: CGConfig, flow_cell_name: str, bcl_converter: str):
    """Add a flow cell to the cgstats database."""
    stats_api: StatsAPI = context.cg_stats_api
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    flow_cell_run_path: Path = Path(demultiplex_api.run_dir, flow_cell_name)
    if not flow_cell_run_path.exists():
        LOG.warning(f"Could not find flow cell path {flow_cell_run_path}")
        raise click.Abort
    demux_results_path: Path = Path(demultiplex_api.out_dir, flow_cell_name)
    if not demux_results_path.exists():
        LOG.warning(f"Could not find demultiplex result path {demux_results_path}")
        raise click.Abort
    try:
        flow_cell: FlowCell = FlowCell(
            flow_cell_path=flow_cell_run_path, bcl_converter=bcl_converter
        )
    except FlowCellError as error:
        raise click.Abort from error
    demux_results: DemuxResults = DemuxResults(
        demux_dir=demux_results_path, flow_cell=flow_cell, bcl_converter=bcl_converter
    )
    create_novaseq_flowcell(manager=stats_api, demux_results=demux_results)


@click.command(name="select")
@click.argument("flow-cell-id")
@click.option("--project", help="Name of project to fetch data for")
@click.pass_obj
def select_project_cmd(context: CGConfig, flow_cell_id: str, project: Optional[str]):
    """Select a flow cell to fetch statistics from"""
    # Need to instantiate API
    post_processing_api = DemuxPostProcessingNovaseqAPI(config=context)
    report_content: List[str] = post_processing_api.get_report_data(
        flow_cell_id=flow_cell_id, project_name=project
    )

    click.echo("\t".join(STATS_HEADER))
    if not report_content:
        LOG.warning(f"Could not find any samples for flow cell {flow_cell_id}, project {project}")
        return
    for report_line in report_content:
        click.echo(report_line)



import logging
from typing import List

from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cg.apps.lims.samplesheet import LimsFlowcellSample
from cg.exc import FlowCellError
from cg.models.demultiplex.flow_cell import FlowCell
from cg.models.demultiplex.run_parameters import RunParameters

LOG = logging.getLogger(__name__)


def create_sample_sheet(
    bcl_converter: str,
    flow_cell: FlowCell,
    lims_samples: List[LimsFlowcellSample],
    force: bool = False,
) -> str:
    """Create a sample sheet for a flow cell."""
    if flow_cell.sample_sheet_path.exists():
        message = f"Sample sheet {flow_cell.sample_sheet_path} already exists!"
        LOG.warning(message)
        raise FileExistsError(message)

    run_parameters: RunParameters = flow_cell.run_parameters_object

    if run_parameters.flowcell_type != "novaseq":
        message = f"Can only demultiplex novaseq with cg. Found type {run_parameters.flowcell_type}"
        LOG.warning(message)
        raise FlowCellError(message=message)

    sample_sheet_creator = SampleSheetCreator(
        bcl_converter=bcl_converter,
        flowcell_id=flow_cell.id,
        lims_samples=lims_samples,
        run_parameters=run_parameters,
        force=force,
    )
    return sample_sheet_creator.construct_sample_sheet()



"""Fixtures for the upload Scout API tests."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Generator, List

import pytest
from housekeeper.store.models import Version

from cg.constants import DataDelivery, FileExtensions, Pipeline
from cg.constants.constants import FileFormat, PrepCategory
from cg.constants.sequencing import SequencingMethod
from cg.io.controller import ReadFile
from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.models.scout.scout_load_config import MipLoadConfig
from cg.store import Store
from cg.store.models import Analysis, Family

# Mocks
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.madeline import MockMadelineAPI
from tests.mocks.mip_analysis_mock import MockMipAnalysis
from tests.mocks.scout import MockScoutAPI
from tests.store_helpers import StoreHelpers

LOG = logging.getLogger(__name__)
SNV_FILE_NAME: str = f"snv{FileExtensions.VCF}"


@pytest.fixture(name="rna_case_id")
def fixture_rna_case_id() -> str:
    """Return a rna case id."""
    return "affirmativecat"


@pytest.fixture(name="dna_case_id")
def fixture_dna_case_id(case_id: str) -> str:
    """Return a DNA case id."""
    return case_id


@pytest.fixture(name="rna_sample_son_id")
def fixture_rna_sample_son_id() -> str:
    """Return a son RNA sample id."""
    return "rna_son"


@pytest.fixture(name="rna_sample_daughter_id")
def fixture_rna_sample_daughter_id() -> str:
    """Return a daughter RNA sample id."""
    return "rna_daughter"


@pytest.fixture(name="rna_sample_mother_id")
def fixture_rna_sample_mother_id() -> str:
    """Return a mother RNA sample id."""
    return "rna_mother"


@pytest.fixture(name="rna_sample_father_id")
def fixture_rna_sample_father_id() -> str:
    """Return a father RNA sample id."""
    return "rna_father"


@pytest.fixture(name="dna_sample_son_id")
def fixture_dna_sample_son_id() -> str:
    """Return a son DNA sample id."""
    return "dna_son"


@pytest.fixture(name="dna_sample_daughter_id")
def fixture_dna_sample_daughter_id() -> str:
    """Return a daughter DNA sample id."""
    return "dna_daughter"


@pytest.fixture(name="dna_sample_mother_id")
def fixture_dna_sample_mother_id() -> str:
    """Return a mother DNA sample id."""
    return "dna_mother"


@pytest.fixture(name="dna_sample_father_id")
def fixture_dna_sample_father_id() -> str:
    """Return a father DNA sample id."""
    return "dna_father"


@pytest.fixture(name="another_sample_id")
def fixture_another_sample_id() -> str:
    """Return another sample id."""
    return "another_sample_id"


@pytest.fixture(name="another_rna_sample_id")
def fixture_another_rna_sample_id() -> str:
    """Return another RNA sample id."""
    return "another_rna_sample_id"


@pytest.fixture(name="rna_store")
def fixture_rna_store(
    base_store: Store,
    helpers: StoreHelpers,
    rna_case_id: str,
    dna_case_id: str,
) -> Store:
    """Populate store with a RNA case that is connected to a DNA case via sample.subject_id."""

    store: Store = base_store

    # an existing RNA case with related sample
    rna_case = helpers.ensure_case(
        store=store,
        case_name="rna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Pipeline.MIP_RNA,
        data_delivery=DataDelivery.SCOUT,
    )
    rna_case.internal_id = rna_case_id

    rna_sample_son = helpers.add_sample(
        store=store,
        name="rna_son",
        subject_id="son",
        application_type=SequencingMethod.WTS,
    )
    rna_sample_daughter = helpers.add_sample(
        store=store,
        name="rna_daughter",
        subject_id="daughter",
        application_type=SequencingMethod.WTS,
    )
    rna_sample_mother = helpers.add_sample(
        store=store,
        name="rna_mother",
        subject_id="mother",
        application_type=SequencingMethod.WTS,
    )
    rna_sample_father = helpers.add_sample(
        store=store,
        name="rna_father",
        subject_id="father",
        application_type=SequencingMethod.WTS,
    )
    helpers.add_relationship(
        store=store,
        sample=rna_sample_son,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status="affected",
    )
    helpers.add_relationship(
        store=store,
        sample=rna_sample_daughter,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status="unaffected",
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_mother, case=rna_case, status="unaffected"
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_father, case=rna_case, status="affected"
    )

    for link in rna_case.links:
        link.sample.internal_id = link.sample.name

    # an existing DNA case with related sample
    dna_case = helpers.ensure_case(
        store=store,
        case_name="dna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Pipeline.MIP_DNA,
        data_delivery=DataDelivery.SCOUT,
    )
    dna_case.internal_id = dna_case_id

    dna_sample_son = helpers.add_sample(
        store=store,
        name="dna_son",
        subject_id="son",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    dna_sample_daughter = helpers.add_sample(
        store=store,
        name="dna_daughter",
        subject_id="daughter",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    dna_sample_mother = helpers.add_sample(
        store=store,
        name="dna_mother",
        subject_id="mother",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    dna_sample_father = helpers.add_sample(
        store=store,
        name="dna_father",
        subject_id="father",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    helpers.add_relationship(
        store=store,
        sample=dna_sample_son,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status="affected",
    )
    helpers.add_relationship(
        store=store,
        sample=dna_sample_daughter,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status="unaffected",
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_mother, case=dna_case, status="unaffected"
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_father, case=dna_case, status="affected"
    )

    for link in dna_case.links:
        link.sample.internal_id = link.sample.name

    store.session.commit()
    return store


@pytest.fixture(name="lims_family")
def fixture_lims_family(fixtures_dir) -> dict:
    """Returns a LIMS-like case of samples."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(fixtures_dir, "report", "lims_family.json")
    )


@pytest.fixture(name="lims_samples")
def fixture_lims_samples(lims_family: dict) -> List[dict]:
    """Returns the samples of a LIMS case."""
    return lims_family["samples"]


@pytest.fixture(scope="function", name="mip_dna_analysis_hk_bundle_data")
def fixture_mip_dna_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, mip_dna_analysis_dir: Path, sample_id: str
) -> dict:
    """Return MIP DNA bundle data for Housekeeper."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": Path(mip_dna_analysis_dir, SNV_FILE_NAME).as_posix(),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "sv.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "snv_research.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-snv-research"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "sv_research.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-research"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "str.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-str"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "smn.vcf").as_posix(),
                "archive": False,
                "tags": ["smn-calling"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "adm1.cram").as_posix(),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "report.pdf").as_posix(),
                "archive": False,
                "tags": ["delivery-report"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "adm1.mt.bam").as_posix(),
                "archive": False,
                "tags": ["bam-mt", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "vcf2cytosure.txt").as_posix(),
                "archive": False,
                "tags": ["vcf2cytosure", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "multiqc.html").as_posix(),
                "archive": False,
                "tags": ["multiqc-html", sample_id],
            },
        ],
    }


@pytest.fixture(scope="function", name="mip_rna_analysis_hk_bundle_data")
def fixture_mip_rna_analysis_hk_bundle_data(
    rna_case_id: str,
    timestamp: datetime,
    mip_dna_analysis_dir: Path,
    rna_sample_son_id: str,
    rna_sample_daughter_id: str,
    rna_sample_mother_id: str,
    rna_sample_father_id: str,
) -> dict:
    """Return MIP RNA bundle data for Housekeeper."""

    files: [dict] = [
        {
            "path": Path(mip_dna_analysis_dir, f"{rna_case_id}_report.selected.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "clinical", rna_case_id],
        },
        {
            "path": Path(mip_dna_analysis_dir, f"{rna_case_id}_report.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "research", rna_case_id],
        },
    ]
    for sample_id in [
        rna_sample_son_id,
        rna_sample_daughter_id,
        rna_sample_mother_id,
        rna_sample_father_id,
    ]:
        files.extend(
            [
                {
                    "path": Path(
                        mip_dna_analysis_dir, f"{sample_id}_lanes_1_star_sorted_sj.bigWig"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["coverage", "bigwig", "scout", sample_id],
                },
                {
                    "path": Path(
                        mip_dna_analysis_dir, f"{sample_id}_lanes_1234_star_sorted_sj.bed.gz.tbi"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["bed", "scout", "junction", sample_id],
                },
            ]
        )

    return {
        "name": rna_case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": files,
    }


@pytest.fixture(scope="function", name="balsamic_analysis_hk_bundle_data")
def fixture_balsamic_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, balsamic_wgs_analysis_dir: Path, sample_id: str
) -> dict:
    """Return Balsamic bundle data for Housekeeper,"""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": Path(balsamic_wgs_analysis_dir, SNV_FILE_NAME).as_posix(),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "sv.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "umi.sv.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-umi-snv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "adm1.cram").as_posix(),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "ascat.output.pdf").as_posix(),
                "archive": False,
                "tags": ["ascatngs", "visualization", sample_id],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "coverage_qc_report.pdf").as_posix(),
                "archive": False,
                "tags": ["delivery-report"],
            },
        ],
    }


@pytest.fixture(scope="function", name="rnafusion_analysis_hk_bundle_data")
def fixture_rnafusion_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, rnafusion_analysis_dir: Path
) -> dict:
    """Get some bundle data for housekeeper."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(rnafusion_analysis_dir / "multiqc.html"),
                "archive": False,
                "tags": ["multiqc-html", "rna"],
            },
            {
                "path": Path(rnafusion_analysis_dir, "rnafusion_report.html").as_posix(),
                "archive": False,
                "tags": ["fusionreport", "research"],
            },
        ],
    }


@pytest.fixture(name="balsamic_analysis_hk_version")
def fixture_balsamic_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return Housekeeper version for a Balsamic bundle."""
    return helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)


@pytest.fixture(name="mip_dna_analysis_hk_version")
def fixture_mip_dna_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return Housekeeper version for a MIP DNA bundle."""
    return helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)


@pytest.fixture(name="mip_dna_analysis_hk_api")
def fixture_mip_dna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with MIP DNA analysis files."""
    helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="mip_rna_analysis_hk_api")
def fixture_mip_rna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_rna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with MIP RNA analysis files."""
    helpers.ensure_hk_version(housekeeper_api, mip_rna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="balsamic_analysis_hk_api")
def fixture_balsamic_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with Balsamic analysis files."""
    helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="rnafusion_analysis_hk_api")
def fixture_rnafusion_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, rnafusion_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some rnafusion analysis files"""
    helpers.ensure_hk_version(housekeeper_api, rnafusion_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="mip_file_handler")
def fixture_mip_file_handler(mip_dna_analysis_hk_version: Version) -> MipConfigBuilder:
    """Return a MIP confiig builder."""
    return MipConfigBuilder(hk_version_obj=mip_dna_analysis_hk_version)


@pytest.fixture(name="mip_dna_analysis")
def fixture_mip_dna_analysis(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers: StoreHelpers
) -> Analysis:
    """Return a MIP DNA analysis object."""
    helpers.add_synopsis_to_case(store=analysis_store_trio, case_id=case_id)
    case: Family = analysis_store_trio.get_case_by_internal_id(internal_id=case_id)
    analysis: Analysis = helpers.add_analysis(
        store=analysis_store_trio,
        case=case,
        started_at=timestamp,
        pipeline=Pipeline.MIP_DNA,
        completed_at=timestamp,
    )
    for link in case.links:
        helpers.add_phenotype_groups_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_phenotype_terms_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_subject_id_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
    return analysis


@pytest.fixture(name="balsamic_analysis_obj")
def fixture_balsamic_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a Balsamic analysis object."""
    analysis_obj.pipeline = Pipeline.BALSAMIC
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = (
            PrepCategory.WHOLE_EXOME_SEQUENCING
        )
        link_object.family.data_analysis = Pipeline.BALSAMIC
    return analysis_obj


@pytest.fixture(name="balsamic_umi_analysis_obj")
def fixture_balsamic_umi_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a Balsamic UMI analysis object."""
    analysis_obj.pipeline = Pipeline.BALSAMIC_UMI
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = (
            PrepCategory.WHOLE_EXOME_SEQUENCING
        )
        link_object.family.data_analysis = Pipeline.BALSAMIC_UMI

    return analysis_obj


@pytest.fixture(name="rnafusion_analysis_obj")
def fixture_rnafusion_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a RNAfusion analysis object."""
    analysis_obj.pipeline = Pipeline.RNAFUSION
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = (
            PrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING
        )
        link_object.family.data_analysis = Pipeline.RNAFUSION
        link_object.family.panels = None
    return analysis_obj


@pytest.fixture(name="mip_config_builder")
def fixture_mip_config_builder(
    mip_dna_analysis_hk_version: Version,
    mip_dna_analysis: Analysis,
    lims_api: MockLimsAPI,
    mip_analysis_api: MockMipAnalysis,
    madeline_api: MockMadelineAPI,
) -> MipConfigBuilder:
    """Return a MIP config builder."""
    return MipConfigBuilder(
        hk_version_obj=mip_dna_analysis_hk_version,
        analysis_obj=mip_dna_analysis,
        lims_api=lims_api,
        mip_analysis_api=mip_analysis_api,
        madeline_api=madeline_api,
    )


@pytest.fixture(name="balsamic_config_builder")
def fixture_balsamic_config_builder(
    balsamic_analysis_hk_version: Version,
    balsamic_analysis_obj: Analysis,
    lims_api: MockLimsAPI,
) -> BalsamicConfigBuilder:
    """Return a Balsamic config builder."""
    return BalsamicConfigBuilder(
        hk_version_obj=balsamic_analysis_hk_version,
        analysis_obj=balsamic_analysis_obj,
        lims_api=lims_api,
    )


@pytest.fixture(name="mip_load_config")
def fixture_mip_load_config(
    mip_dna_analysis_dir: Path, case_id: str, customer_id: str
) -> MipLoadConfig:
    """Return a valid MIP load_config."""
    return MipLoadConfig(
        owner=customer_id,
        family=case_id,
        vcf_snv=Path(mip_dna_analysis_dir, SNV_FILE_NAME).as_posix(),
        track="rare",
    )


@pytest.fixture(name="lims_api")
def fixture_lims_api(lims_samples: List[dict]) -> MockLimsAPI:
    """Return a LIMS API."""
    return MockLimsAPI(samples=lims_samples)


@pytest.fixture(name="mip_analysis_api")
def fixture_mip_analysis_api() -> MockMipAnalysis:
    """Return a MIP analysis API."""
    return MockMipAnalysis()


@pytest.fixture(name="upload_scout_api")
def fixture_upload_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    housekeeper_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Return upload Scout API."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    return UploadScoutAPI(
        hk_api=housekeeper_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture(name="upload_mip_analysis_scout_api")
def fixture_upload_mip_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    mip_dna_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return MIP upload Scout API."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=mip_dna_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture(name="upload_balsamic_analysis_scout_api")
def fixture_upload_balsamic_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    balsamic_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return Balsamic upload Scout API."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=balsamic_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture(name="upload_rnafusion_analysis_scout_api")
def fixture_upload_rnafusion_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    rnafusion_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    _api = UploadScoutAPI(
        hk_api=rnafusion_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )

    return _api
